#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of galpynostatic
#   https://github.com/fernandezfran/galpynostatic/
# Copyright (c) 2022-2023, Francisco Fernandez
# License: MIT
#   https://github.com/fernandezfran/galpynostatic/blob/master/LICENSE

# ============================================================================
# DOCS
# ============================================================================

r"""The ``galpynostatic.dataset`` module loads the data required for the fits.

These datasets have been obtained using a continuum computational physics model
for different geometries [5]_. They result from a cut-off of multiple
galvanostatic profiles at a given cell potential, with respect to equilibrium,
with different combinations of the internal unitless parameters :math:`\Xi` and
:math:`\ell` covering a wide range of possible values of the experimental
descriptors involved.

References
----------
.. [5] E. M. Gavilán-Arriazu, D. E. Barraco, D., Y. Ein-Eli and E. P. M. Leiva.
   "Galvanostatic Fast Charging of Alkali‐Ion Battery Materials at the
   Single‐Particle Level: A Map‐Driven Diagnosis.i" `ChemPhysChem` 24, no. 6
   (2023): e202200665.
"""

# ============================================================================
# IMPORTS
# ============================================================================

import os
import pathlib

import pandas as pd

# ============================================================================
# CONSTANTS
# ============================================================================

PATH = pathlib.Path(os.path.abspath(os.path.dirname(__file__)))

# ============================================================================
# FUNCTIONS
# ============================================================================


def load_dataset(geometry="spherical"):
    """Galvanostatic map for a cut-off potential of 150 mV.

    Parameters
    ----------
    geometry : str, default="spherical"
        The geometry of the electrode. It can be `"spherical"`, `"cylindrical"`
        or `"planar"`.

    Returns
    -------
    pandas.DataFrame
        The geometry dataset as a ``pandas.DataFrame``.

    Raises
    ------
    ValueError
        If the geometry is not `"spherical"`, `"cylindrical"` or `"planar"`.
    """
    try:
        return pd.read_csv(PATH / f"{geometry}.csv")
    except FileNotFoundError:
        raise ValueError(f"{geometry} is not a valid geometry.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of galpynostatic
#   https://github.com/fernandezfran/galpynostatic/
# Copyright (c) 2022-2023, Francisco Fernandez
# License: MIT
#   https://github.com/fernandezfran/galpynostatic/blob/master/LICENSE

# ============================================================================
# DOCS
# ============================================================================

"""Module with the galvanostatic regression model."""

# ============================================================================
# IMPORTS
# ============================================================================

import itertools as it

import numpy as np

import pandas as pd

from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.metrics import mean_squared_error as _skl_mse
from sklearn.utils import validation as _skl_validation

from .base import MapSpline
from .datasets import load_dataset
from .plot import GalvanostaticPlotter
from .utils import logell, logxi

# ============================================================================
# CLASSES
# ============================================================================


class GalvanostaticRegressor(BaseEstimator, RegressorMixin):
    r"""A heuristic regressor for State-of-Charge (SOC) versus C-rates data.

    This physics-based heuristic model [4]_ uses the maps in the
    :ref:`galpynostatic.datasets` to perform a grid search, using different
    combinations of the diffusion coefficient, :math:`D`, and the
    kinetic rate constant, :math:`k^0`, to fit experimental data of the
    State-of-Charge (SOC) of the electrode material as a function of the
    galvanostatic charging rate (C-rate). This is done while keeping invariant
    all the other experimental descriptors involved in the parameters
    :math:`\Xi` and :math:`\ell` of the continuum model [4]_, such
    as the characteristic diffusion length, :math:`d`, and the geometric
    factor, :math:`z` (see :ref:`galpynostatic.utils`).

    Each time a set of parameters :math:`D` and :math:`k^0` are taken, the
    SOC values are predicted and the mean square error (MSE) is calculated.
    The set of parameters that minimises the MSE is then obtained, providing
    a fundamental description of the system.

    Parameters
    ----------
    dataset : str or pandas.DataFrame, default="spherical"
        A str specifying the particle geometry (`"planar"`, `"cylindrical"` or
        `"spherical"`) to use the datasets distributed in this package, which
        can also be loaded using the functions of the
        :ref:`galpynostatic.datasets` to get it as a ``pandas.DataFrame`` with
        the mapping of maximum SOC values as a function of the internal
        parameters :math:`\log(\ell)` and :math:`\log(\Xi)`.

    d : float, default=1e-4
        Characteristic diffusion length (particle size) in cm.

    z : int, default=3
        Geometric factor (`1` for planar, `2` for cylinder and `3` for sphere).

    dcoeff_lle : int, default=-15
        The lower limit exponent of the diffusion coefficient line used to
        generate the grid.

    dcoeff_ule : int, default=-6
        The upper limit exponent of the diffusion coefficient line used to
        generate the grid.

    dcoeff_num : int, default=100
        Number of samples of diffusion coefficients to generate between the
        lower and the upper limit exponents.

    k0_lle : int, default=-14
        The lower limit exponent of the kinetic rate constant line used to
        generate the grid.

    k0_ule : int, default=-5
        The upper limit exponent of the kinetic rate constant line used to
        generate the grid.

    k0_num : int, default=100
        Number of samples of kinetic rate constants to generate between the
        lower and the upper limit exponents.

    Notes
    -----
    You can also give your own dataset to another potential cut-off in the
    same format as the distributed ones and as ``pandas.DataFrame``, i.e. in
    the column of :math:`\ell` the different values have to be grouped in
    ascending order and for each of these groups the :math:`\Xi` have to be in
    descending order and respecting that for each group of :math:`\ell` the
    same values are simulated (this is a restriction to perform the
    ``scipy.interpolate.RectBivariateSpline``, since `x` and `y` must be
    strictly in a special order, which is handled internally by the
    :ref:`galpynostatic.base`).

    References
    ----------
    .. [4] F. Fernandez, E. M. Gavilán-Arriazu, D. E. Barraco, A. Visintin, Y.
       Ein-Eli and E. P. M. Leiva. "Towards a fast-charging of LIBs electrode
       materials: a heuristic model based on galvanostatic simulations."
       `Electrochimica Acta 464` (2023): 142951.

    Attributes
    ----------
    dcoeff_ : float
        Predicted diffusion coefficient in :math:`cm^2/s`.

    dcoeff_err_ : float
        Uncertainty of the predicted diffusion coefficient.

    k0_ : float
        Predicted kinetic rate constant in :math:`cm/s`.

    k0_err_ : float
        Uncertainty of the predicted kinetic rate constant.

    mse_ : float
        Mean squared error of the best fitted model.
    """

    def __init__(
        self,
        dataset="spherical",
        d=1e-4,
        z=3,
        dcoeff_lle=-15,
        dcoeff_ule=-6,
        dcoeff_num=100,
        k0_lle=-14,
        k0_ule=-5,
        k0_num=100,
    ):
        self.dataset = dataset
        self.d = d
        self.z = z
        self.dcoeff_lle = dcoeff_lle
        self.dcoeff_ule = dcoeff_ule
        self.dcoeff_num = dcoeff_num
        self.k0_lle = k0_lle
        self.k0_ule = k0_ule
        self.k0_num = k0_num

    def _validate_geometry(self):
        """Validate geometry (when dataset is a string)."""
        if isinstance(self.dataset, str):
            self.dataset = load_dataset(geometry=self.dataset)

    def _grid_points(self):
        """Grid points (D, k0) to evaluate in the grid search."""
        dcoeffs = np.logspace(
            self.dcoeff_lle, self.dcoeff_ule, num=self.dcoeff_num
        )
        k0s = np.logspace(self.k0_lle, self.k0_ule, num=self.k0_num)
        return np.array(tuple(it.product(dcoeffs, k0s)))

    def _calculate_uncertainties(self, X, y, attrs, delta):
        """Uncertainties of `attrs` calculation.

        The uncertainties are computed as the root squared values of the
        diagonal in the covariance matrix, which is approximated with the
        inverse of the Hessian matrix, calculated as the product of the
        Jacobian matrix with its transpose.
        """
        jacobian = np.zeros((len(attrs), len(y)))
        for i, attr in enumerate(attrs):
            param = self.__dict__[attr]

            self.__dict__[attr] = (1 + delta) * param
            upper = self.predict(X)

            self.__dict__[attr] = (1 - delta) * param
            lower = self.predict(X)

            jacobian[i] = (upper - lower) / (2 * delta * param)

        hessian = np.dot(jacobian, jacobian.T)
        covariance = np.linalg.inv(hessian)

        stdsq = np.sum((y - self.predict(X)) ** 2) / (len(y) - len(attrs))

        return stdsq * np.sqrt(np.diag(covariance))

    def fit(self, X, y, sample_weight=None):
        """Fit the heuristic galvanostatic regressor model.

        Parameters
        ----------
        X : array-like of shape (n_measurements, 1)
            C-rates data of the experiments.

        y : array-like of shape (n_measurements,)
            Target maximum SOC values, between 0 and 1.

        sample_weight : array-like of shape(n_measurements,), default=None
            Individual weights of each data point.

        Returns
        -------
        self : object
            Fitted model.

        Raises
        ------
        ValueError
            If the instantiated dataset is an str but is not a valid geometry
            (`"planar"`, `"cylindrical"` or `"spherical"`).
        """
        X, y = _skl_validation.check_X_y(X, y)
        self._validate_geometry()

        self._map = MapSpline(self.dataset)

        params = self._grid_points()

        mse = np.full(params.shape[0], np.inf)
        for k, (self.dcoeff_, self.k0_) in enumerate(params):
            try:
                mse[k] = _skl_mse(
                    y, self.predict(X), sample_weight=sample_weight
                )
            except ValueError:
                ...

        idx = np.argmin(mse)
        self.mse_ = mse[idx]

        self.dcoeff_, self.k0_ = params[idx]

        self.dcoeff_err_, self.k0_err_ = self._calculate_uncertainties(
            X, y, ("dcoeff_", "k0_"), np.cbrt(np.finfo(float).eps)
        )

        return self

    def predict(self, X):
        """Predict using the heuristic model within the map constraints.

        Parameters
        ----------
        X : array-like of shape (n_measurements, 1)
            C-rates input points.

        Returns
        -------
        y : array-like of shape (n_measurements,)
            The predicted maximum SOC values for the C-rate inputs.
        """
        _skl_validation.check_is_fitted(self)
        X = _skl_validation.check_array(X)

        logells = logell(X.ravel(), self.d, self.z, self.dcoeff_)
        logxis = logxi(X.ravel(), self.dcoeff_, self.k0_)

        mask_logell = self._map._mask_logell(logells)
        mask_logxi = self._map._mask_logxi(logxis)

        return np.where(
            mask_logell & mask_logxi, self._map.soc(logells, logxis), np.nan
        )

    def score(self, X, y, sample_weight=None):
        r"""Return the coefficient of determination of the prediction.

        The coefficient of determination :math:`R^2` is defined as
        :math:`(1 - \frac{u}{v})`, where :math:`u` is the residual
        sum of squares ``((y_experimental - y_pred)** 2).sum()`` and :math:`v`
        is the total sum of squares
        ``((y_experimental - y_experimental.mean()) ** 2).sum()``. The best
        possible score is 1.0 and it can be negative (because the model can be
        arbitrarily worse).

        Parameters
        ----------
        X : array-like of shape (n_measurements, 1)
            C-rates data of the experiments.

        y : array-like of shape (n_measurements,)
            Maximum SOC values of the experiments.

        sample_weight : array-like of shape(n_measurements,), default=None
            Individual weights of each data point.

        Returns
        -------
        score : float
            :math:`R^2` of ``self.predict(X)`` wrt. `y`.
        """
        return super(GalvanostaticRegressor, self).score(
            X, y, sample_weight=sample_weight
        )

    def to_dataframe(self, X, y=None):
        """Convert the train, the evaluation or both sets into a dataframe.

        Get a dataframe with two or three columns (`C_rates`, `SOC_exp` and
        `SOC_pred`), depending on whether you have passed the `y`-values or
        not.

        Parameters
        ----------
        X : array-like of shape (n_measurements, 1)
            C-rate points.

        y : array-like of shape (n_measurements,), default=None
            Maximum SOC values.

        Returns
        -------
        df : pandas.DataFrame
            A ``pandas.DataFrame`` containing the train, the evaluation or both
            sets.
        """
        df = pd.DataFrame({"C_rates": X.ravel()})

        if y is not None:
            df["SOC_exp"] = y

        df["SOC_pred"] = self.predict(X)

        return df

    @property
    def plot(self):
        """Plot accessor for the :ref:`galpynostatic.plot`."""
        return GalvanostaticPlotter(self)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of galpynostatic
#   https://github.com/fernandezfran/galpynostatic/
# Copyright (c) 2022-2023, Francisco Fernandez
# License: MIT
#   https://github.com/fernandezfran/galpynostatic/blob/master/LICENSE

# ============================================================================
# DOCS
# ============================================================================

"""Plot helper for GalvanostaticRegressor object."""

# ============================================================================
# IMPORTS
# ============================================================================

import matplotlib.pyplot as plt

import numpy as np

from .utils import logell, logxi

# ============================================================================
# CLASSES
# ============================================================================


class GalvanostaticPlotter:
    r"""GalvanostaticRegressor plot utilities.

    Type of plots to produce:

    - 'render_map' : the map on which the data have been fitted.
    - 'in_render_map' : :math:`\Xi` and :math:`\ell` data points in the map.
    - 'versus_data' : predicted and actual maximum State-of-Charge (SOC) values
      versus galvanostatic charging rate (C-rate).

    Parameters
    ----------
    greg : galpynostatic.model.GalvanostaticRegressor
        An already fitted GalvanostaticRegressor model.

    Notes
    -----
    The map is only plotted in ``self.in_render_map(X)`` if ax is None,
    otherwise it is assumed to be already plotted and you just want to add
    the points to it, e.g. to compare different systems.
    """

    def __init__(self, greg):
        self.greg = greg

    def render_map(self, ax=None, clb=True, clb_label="maximum SOC"):
        """Plot the map to which the data has been fitted.

        Parameters
        ----------
        ax : matplotlib.axes.Axes, default=None
            The current axes.

        clb: bool, default=True
            Add the colorbar to the plot.

        clb_label : str, default="maximum SOC"
            The label for the colorbar.

        Returns
        -------
        ax : matplotlib.axes.Axes
            The current axes.
        """
        ax = plt.gca() if ax is None else ax

        logelleval = np.linspace(
            np.min(self.greg._map.logells_),
            np.max(self.greg._map.logells_),
            num=1000,
        )
        logxieval = np.linspace(
            np.min(self.greg._map.logxis_),
            np.max(self.greg._map.logxis_),
            num=1000,
        )

        z = self.greg._map.soc(logelleval, logxieval, grid=True)

        im = ax.imshow(
            z.T,
            extent=[
                logelleval.min(),
                logelleval.max(),
                logxieval.min(),
                logxieval.max(),
            ],
            origin="lower",
        )

        if clb:
            clb = plt.colorbar(im)
            clb.ax.set_ylabel(clb_label)
            clb.ax.set_ylim((0, 1))

        ax.set_xlabel(r"log($\ell$)")
        ax.set_ylabel(r"log($\Xi$)")

        return ax

    def in_render_map(self, X, ax=None, **kwargs):
        """Plot showing in which region of the map the fit is found.

        Parameters
        ----------
        X : array-like of shape (n_measurements, 1)
            C-rate data from the experiments.

        ax : matplotlib.axes.Axes, default=None
            The current matplotlib axes.

        **kwargs
            Additional keyword arguments that are passed and are documented in
            ``matplotlib.axes.Axes.plot``

        Returns
        -------
        ax : matplotlib.axes.Axes
            The current axes.
        """
        ax = self.render_map() if ax is None else ax

        keys = ["color", "marker", "linestyle", "label"]
        for key, value in zip(keys, ["k", "o", "--", "fitted data"]):
            kwargs.setdefault(key, value)

        ax.plot(
            logell(X.ravel(), self.greg.d, self.greg.z, self.greg.dcoeff_),
            logxi(X.ravel(), self.greg.dcoeff_, self.greg.k0_),
            **kwargs,
        )

        return ax

    def versus_data(
        self, X, y, X_eval=None, ax=None, data_kws=None, pred_kws=None
    ):
        """Plot SOC predictions versus target data as a function of the C-rate.

        Parameters
        ----------
        X : array-like of shape (n_measurements, 1)
            C-rates data of the experiments.

        y : array-like of shape (n_measurements,)
            Target maximum SOC values.

        X_eval : array-like of shape (n_measurements, 1), default=None.
            C-rates values to evaluate the model and compare it with the data.
            If set to `None`, it will evaluate 250 points between the maximum
            and minimum of X.

        ax : matplotlib.axes.Axes, default=None
            The current axes.

        data_kws : dict, default=None
            Additional keyword arguments that are passed and are documented in
            ``matplotlib.axes.Axes.plot`` for the data points.

        pred_kws : dict, default=None
            Additional keyword arguments that are passed and are documented in
            ``matplotlib.axes.Axes.plot`` for the prediction values.

        Returns
        -------
        ax : matplotlib.axes.Axes
            The current axes.
        """
        ax = plt.gca() if ax is None else ax

        data_kws = {} if data_kws is None else data_kws
        pred_kws = {} if pred_kws is None else pred_kws

        keys = ["color", "marker", "linestyle", "label"]

        for key, value in zip(keys, ["tab:blue", "s", "--", "data"]):
            data_kws.setdefault(key, value)

        for key, value in zip(keys, ["tab:orange", "", "-", "model"]):
            pred_kws.setdefault(key, value)

        ax.plot(X, y, **data_kws)

        X_eval = (
            np.linspace(X.min(), X.max(), 250).reshape(-1, 1)
            if X_eval is None
            else X_eval
        )
        ax.plot(X_eval, self.greg.predict(X_eval), **pred_kws)

        ax.set_xlabel("C-rates")
        ax.set_ylabel("maximum SOC")

        ax.set_xscale("log")

        return ax



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of galpynostatic
#   https://github.com/fernandezfran/galpynostatic/
# Copyright (c) 2022-2023, Francisco Fernandez
# License: MIT
#   https://github.com/fernandezfran/galpynostatic/blob/master/LICENSE

# ============================================================================
# DOCS
# ============================================================================

"""Internal unitless parameters functions."""

# ============================================================================
# IMPORTS
# ============================================================================

import numpy as np

# ============================================================================
# FUNCTIONS
# ============================================================================


def logell(c_rate, d, z, dcoeff):
    r"""Obtain log value in base 10 of :math:`\ell` internal parameter.

    :math:`\ell = \frac{d^2 C_{rate}}{z t_h D}` where :math:`d` is the
    particle size, :math:`C_{rate}` is the galvanostatic charging rate,
    :math:`z` is a geometrical factor, :math:`D` is the diffusion coefficient
    and :math:`t_h` is the time equivalent to one hour in suitable time units,
    here 3600 seconds.

    Parameters
    ----------
    c_rate : float or int or array-like
        C-rate values.

    d : float
        Characteristic diffusion length (particle size) in :math:`cm`.

    z : int
        Geometric factor: 1 for planar, 2 for cylinder and 3 for sphere.

    dcoeff : float
        Diffusion coefficient, :math:`D`, in :math:`cm^2/s`.

    Returns
    -------
    logell : float or array-like
        The log 10 value of :math:`\ell` internal parameter.
    """
    return np.log10((np.asarray(c_rate) * d**2) / (3600 * z * dcoeff))


def logxi(c_rate, dcoeff, k0):
    r"""Obtain log value in base 10 of :math:`\Xi` internal parameter.

    :math:`\Xi = k^0 \sqrt{\frac{t_h}{C_{rate} D}}` where :math:`k^0` is the
    kinetic rate constant, :math:`t_h` is the time equivalent to one hour in
    suitable time units, here 3600 seconds, :math:`C_{rate}` is the
    galvanostatic charging rate and :math:`D` is the diffusion coefficient.

    Parameters
    ----------
    c_rate : float or int or array-like
        C-rate values.

    dcoeff : float
        Diffusion coefficient, :math:`D`, in :math:`cm^2/s`.

    k0 : float
        Kinetic rate constant, :math:`k^0`, in :math:`cm/s`.

    Returns
    -------
    logxi : float or array-like
        The log 10 value of :math:`\Xi` internal parameter.
    """
    return np.log10(k0 * np.sqrt(3600 / (np.asarray(c_rate) * dcoeff)))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of galpynostatic
#   https://github.com/fernandezfran/galpynostatic/
# Copyright (c) 2022-2023, Francisco Fernandez
# License: MIT
#   https://github.com/fernandezfran/galpynostatic/blob/master/LICENSE

# ============================================================================
# DOCS
# ============================================================================

"""Make predictions using the physics-based heuristic model.

The physics-based heuristic model [1]_ presented in this package allows to fit
State-of-Charge (SOC) battery data as a function of galvanostatic charging
rate (C-rate). Once this model has been fitted, the diffusion coefficient,
:math:`D`, and the kinetic rate constant, :math:`k^0`, parameters of the active
material in the electrode remain fixed. The other two parameters of the model,
the characteristic diffusion length, :math:`d`, (particle size) and the C-rate
can be varied. In this way, the model can be used to predict both the optimum
particle size for a given C-rate and the optimum charging rate for a given
particle size to achieve a desired maximum SOC.

References
----------
.. [1] F. Fernandez, E. M. Gavilán-Arriazu, D. E. Barraco, A. Visintin, Y.
   Ein-Eli and E. P. M. Leiva. "Towards a fast-charging of LIBs electrode
   materials: a heuristic model based on galvanostatic simulations."
   `Electrochimica Acta 464` (2023): 142951.
"""

# ============================================================================
# IMPORTS
# ============================================================================

import numpy as np

import scipy.optimize

from .utils import logell, logxi

# ============================================================================
# CONSTANTS
# ============================================================================

VALUE_ERROR_MESSAGE = (
    "This system fails to find the desired point given conditions that may be "
    "due to map constraints or input parameters (such as `loaded`, initial "
    "estimate or keyword arguments passed to `scipy.optimize.newton`)."
)

# ============================================================================
# FUNCTIONS
# ============================================================================


def optimal_charging_rate(greg, c0=1.0, loaded=0.8, **kwargs):
    r"""Predict the optimal C-rate to reach a desired SOC.

    The default parameters of this function predict the C-rate required to
    reach the 80% of the electrode charge.

    Parameters
    ----------
    greg : galpynostatic.model.GalvanostaticRegressor
        An already fitted GalvanostaticRegressor.

    c0 : float, default=4.0
        An initial estimate of the optimal charging rate that should be
        somewhere near the actual prediction.

    loaded : float, default=0.8
        Desired maximum SOC, between 0 and 1.

    **kwargs
        Additional keyword arguments that are passed and are documented in
        ``scipy.optimize.newton``.

    Returns
    -------
    c_rate : float
        The optimal galvanostatic charging rate to charge the electrode to the
        desired maximum SOC value.

    c_rate_err : float, optional
        The uncertainty is only returned if `greg.dcoeff_err_` and
        `greg.k0_err_` are both defined.

    Raises
    ------
    ValueError
        If the material does not meet the defined criterion given the input
        parameters or map constraints.
    """

    def objfunc(cr, greg, loaded):
        return greg.predict(np.reshape([cr], (-1, 1)))[0] - loaded

    try:
        c_rate = scipy.optimize.newton(
            objfunc, c0, args=(greg, loaded), **kwargs
        )
    except (RuntimeError, ValueError):
        raise ValueError(VALUE_ERROR_MESSAGE)

    if greg.dcoeff_err_ is None or greg.k0_err_ is None:
        return c_rate

    else:
        optimal_xi = 10.0 ** logxi(c_rate, greg.dcoeff_, greg.k0_)
        frac = greg.k0_ / greg.dcoeff_
        c_rate_err = (60.0 * np.sqrt(frac) / optimal_xi) * np.hypot(
            frac * greg.dcoeff_err_, 2 * greg.k0_err_
        )

        return c_rate, c_rate_err


def optimal_particle_size(
    greg, d0=1e-4, loaded=0.8, c_rate=4.0, cm_to=10_000, **kwargs
):
    r"""Predict the optimal electrode particle size to charge in certain time.

    The default parameters of this function define the criteria of reaching 80%
    of the electrode charge in 15 minutes, which corresponds to a maximum SOC
    of 0.8 and a C-rate of 4C, which is the USABC (`United States
    Advanced Battery Consortium`) standard for fast charging.

    Parameters
    ----------
    greg : galpynostatic.model.GalvanostaticRegressor
        A GalvanostaticRegressor already fitted.

    d0 : float, default=1e-4
        An initial estimate of the optimal particle size rate that should be
        somewhere near the actual prediction.

    loaded : float, default=0.8
        Desired maximum SOC value, between 0 and 1.

    c_rate : int or float, default=4.0
        Desired C-rate to reach the established SOC.

    cm_to : float, default=10000
        A factor to convert from cm to another unit, in the default case to
        microns.

    **kwargs
        Additional keyword arguments that are passed and are documented in
        ``scipy.optimize.newton``.

    Returns
    -------
    particle_size : float
        The optimal particle size to charge the electrode to the desired
        maximum SOC value in the desired time.

    particle_size_err : float, optional
        The uncertainty is only returned if `greg.dcoeff_err_` is defined.

    Raises
    ------
    ValueError
        If the material does not meet the defined criterion given the input
        parameters or map constraints.
    """

    def objfunc(d, greg, c_rate, loaded):
        greg.d = d
        return greg.predict(np.reshape([c_rate], (-1, 1)))[0] - loaded

    try:
        particle_size = np.abs(
            scipy.optimize.newton(
                objfunc, d0, args=(greg, c_rate, loaded), **kwargs
            )
        )
    except (RuntimeError, ValueError):
        raise ValueError(VALUE_ERROR_MESSAGE)

    if greg.dcoeff_err_ is None:
        return cm_to * particle_size

    else:
        optimal_ell = 10.0 ** logell(
            c_rate, particle_size, greg.z, greg.dcoeff_
        )
        factor = np.sqrt((3600 * greg.z * optimal_ell) / c_rate)
        particle_size_err = (
            cm_to * factor * greg.dcoeff_err_ / (2 * np.sqrt(greg.dcoeff_))
        )

        return (cm_to * particle_size, particle_size_err)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of galpynostatic
#   https://github.com/fernandezfran/galpynostatic/
# Copyright (c) 2022-2023, Francisco Fernandez
# License: MIT
#   https://github.com/fernandezfran/galpynostatic/blob/master/LICENSE

# ============================================================================
# DOCS
# ============================================================================

"""Classes common to different modules."""

# ============================================================================
# IMPORTS
# ============================================================================

import numpy as np

import scipy.interpolate

# ============================================================================
# CLASSES
# ============================================================================


class MapSpline:
    r"""Spline of a discrete map dataset.

    Parameters
    ----------
    dataset : pandas.DataFrame
        Dataset with the discrete map data points of the maximum
        State-of-Charge (SOC) values as a function of the internal parameters
        :math:`\log(\ell)` and :math:`\log(\Xi)`, this can be loaded using the
        functions of the :ref:`galpynostatic.datasets`. See the Notes in the
        :ref:`galpynostatic.model` to know the restrictions of these dataframe.

    Attributes
    ----------
    logells_ : numpy.ndarray
        Unique :math:`\ell` values defined in the dataset.

    logxis_ : numpy.ndarray
        Unique :math:`\Xi` values defined in the dataset.

    spline_ : scipy.interpolate.RectBivariateSpline
        Bivariate spline approximation over the discrete dataset.
    """

    def __init__(self, dataset):
        self.logells_ = np.unique(dataset.l)
        self.logxis_ = np.unique(dataset.xi)

        socs = dataset.xmax.to_numpy().reshape(
            self.logells_.size, self.logxis_.size
        )[:, ::-1]

        self.spline_ = scipy.interpolate.RectBivariateSpline(
            self.logells_, self.logxis_, socs
        )

    def _mask_logell(self, logell):
        """Mask the values between the extrems of the interval."""
        return np.logical_and(
            np.greater_equal(logell, self.logells_.min()),
            np.less_equal(logell, self.logells_.max()),
        )

    def _mask_logxi(self, logxi):
        """Mask the values between the extrems of the interval."""
        return np.logical_and(
            np.greater_equal(logxi, self.logxis_.min()),
            np.less_equal(logxi, self.logxis_.max()),
        )

    def soc(self, logell, logxi, grid=False):
        r"""Predicts the maximum SOC values using the map spline.

        This is a linear function of the spline bounded in [0, 1], values
        outside this range are taken to the corresponding endpoint.

        Parameters
        ----------
        logell : numpy.ndarray
            Log 10 value of :math:`\ell` parameter.

        logxi : numpy.ndarray
            Log 10 value of :math:`\Xi` parameter.

        grid : bool, default=False
            Whether to evaluate the results on a grid spanned by the input
            arrays, or at points specified by the input arrays, i.e. True
            is a grid and False only the ordered pairs.

        Returns
        -------
        soc : numpy.ndarray
            The corresponding maximum SOC values in the map spline.
        """
        return np.clip(self.spline_(logell, logxi, grid=grid), 0, 1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of galpynostatic
#   https://github.com/fernandezfran/galpynostatic/
# Copyright (c) 2022-2023, Francisco Fernandez
# License: MIT
#   https://github.com/fernandezfran/galpynostatic/blob/master/LICENSE

# ============================================================================
# DOCS
# ============================================================================

"""Module to handle experimental data preprocessing."""

# ============================================================================
# IMPORTS
# ============================================================================

import numpy as np

import scipy.interpolate

from sklearn.base import TransformerMixin

# ============================================================================
# CLASSES
# ============================================================================


class GetDischargeCapacities(TransformerMixin):
    """Obtain the discharge capacities at a given cut-off potential.

    This Transformer subtracts the equilibrium potential, `eq_pot`,
    from all galvanostatic profiles the to find the discharge capacity for
    each curve where the potential is cut off by `vcut`.

    Parameters
    ----------
    eq_pot : float, default=0.0
        The equilibrium potential in volts (V).

    vcut : float, default=0.15
        The cut-off potential in V. The default value is 150 mV, which is the
        value defined by the distributed map data in the
        :ref:`galpynostatic.datasets`.

    Notes
    -----
    Discharge capacities are useful to define the maximum value of the
    State-of-Charge (SOC) for a given galvanostatic charging rate (C-rate),
    which is the appropriate way to have the data for the
    :ref:`galpynostatic.model`. Our suggestion for determining the maximum
    value of the SOC is to take the maximum value of the discharging
    capacities corresponding to the value of the C-rate to which the curve
    already converges with respect to the previous one. In this case, all the
    values obtained for the discharge capacities are divided by this value and
    the maximum SOC values are obtained.
    """

    def __init__(self, eq_pot=0.0, vcut=0.15):
        self.eq_pot = eq_pot
        self.vcut = vcut

    def fit(self, X, y=None, **fit_params):
        """Define the fit parameters.

        Parameters
        ----------
        X : Ignored
            Not used, presented for API consistency.

        y : Ignored
            Not used, presented by convention for sklearn API consistency.

        **fit_params
            Additional keyword arguments that are passed and are documented in
            ``scipy.interpolate.InterpolatedUnivariateSpline``.
        """
        self.fit_params = fit_params
        return self

    def transform(self, X):
        """Transform the curves to single values of discharge capacities.

        Parameters
        ----------
        X : list of pandas.DataFrame
            DataFrames having only two columns, the first being the capacity
            and the second being the voltage.

        Returns
        -------
        X_new : array-like of shape (n_measurements,)
            Discharge capacities in the same order as the ``pandas.DataFrame``
            in the input list, but reshaped to fit.
        """
        X_new = np.zeros(len(X))

        for k, df in enumerate(X):
            capacity = df.iloc[:, 0]
            voltage = df.iloc[:, 1] - self.eq_pot + self.vcut

            spline = scipy.interpolate.InterpolatedUnivariateSpline(
                capacity, voltage, **self.fit_params
            )

            try:
                X_new[k] = spline.roots()[0]
            except IndexError:
                ...

        return X_new

    def fit_transform(self, X, y=None, **fit_params):
        """Transform the curves to discharge capacities with optional params.

        Parameters
        ----------
        X : list of pandas.DataFrame
            DataFrames having only two columns, the first being the capacity
            and the second being the voltage.

        y : Ignored
            Not used, presented for sklearn API consistency by convention.

        **fit_params
            Additional keyword arguments that are passed and are documented in
            ``scipy.interpolate.InterpolatedUnivariateSpline``.

        Returns
        -------
        X_new : array-like of shape (n_measurements,)
            Discharge capacities in the same order as the ``pandas.DataFrame``
            in the input list, but reshaped to fit.
        """
        return super(GetDischargeCapacities, self).fit_transform(
            X, y, **fit_params
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of galpynostatic
#   https://github.com/fernandezfran/galpynostatic/
# Copyright (c) 2022-2023, Francisco Fernandez
# License: MIT
#   https://github.com/fernandezfran/galpynostatic/blob/master/LICENSE

# =============================================================================
# DOCS
# =============================================================================

"""A Python package with physics-based models to predict optimal conditions \
for fast-charging lithium-ion batteries."""

# =============================================================================
# IMPORTS
# =============================================================================

import importlib_metadata

from . import datasets
from .base import MapSpline
from .make_prediction import optimal_charging_rate, optimal_particle_size
from .metric import bmxfc, fom
from .model import GalvanostaticRegressor
from .preprocessing import GetDischargeCapacities


# =============================================================================
# CONSTANTS
# =============================================================================

__all__ = [
    "bmxfc",
    "datasets",
    "fom",
    "GalvanostaticRegressor",
    "GetDischargeCapacities",
    "MapSpline",
    "optimal_charging_rate",
    "optimal_particle_size",
]


NAME = "galpynostatic"

DOC = __doc__

VERSION = importlib_metadata.version(NAME)

__version__ = tuple(VERSION.split("."))

del importlib_metadata



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of galpynostatic
#   https://github.com/fernandezfran/galpynostatic/
# Copyright (c) 2022-2023, Francisco Fernandez
# License: MIT
#   https://github.com/fernandezfran/galpynostatic/blob/master/LICENSE

# ============================================================================
# DOCS
# ============================================================================

"""Metrics for benchmarking fast charging battery electrode materials."""

# ============================================================================
# IMPORTS
# ============================================================================

import numpy as np

from .base import MapSpline
from .model import GalvanostaticRegressor

# ============================================================================
# FUNCTIONS
# ============================================================================


def bmxfc(greg, minutes=15, loaded=0.8, full_output=False, **kwargs):
    r"""Metric for benchmarking an extreme fast charging of Li-ion materials.

    This universal metric for Benchmarking battery electrode Materials for an
    eXtreme Fast Charging (BMXFC) is defined as the maximum State-of-Charge
    (SOC) retained when a material is charged for 15 minutes under constant
    current conditions [2]_. The evaluation of the BMXFC is performed using the
    model in this package, which accounts for finite diffusion, charge
    transfer, particle size and the total charging rate.

    Parameters
    ----------
    greg : galpynostatic.model.GalvanostaticRegressor or dict
        An already fitted GalvanostaticRegressor model or a dict containing
        the following keys with float values definened: `d` in :math:`cm`
        (particle size), `dcoeff_` in :math:`cm^2/s` (diffusion coefficient)
        and `k0_` in :math:`cm/s` (kinetic rate constant).

    minutes : int or float, default=15
        Minutes of loading.

    loaded : float, default=0.8
        Criteria for considering the electrode material with fast charging
        capabilities.

    full_output : bool, default=False
        If `full_output` is `False` (default case), the SOC value is returned.
        If it is `True` a dict is returned with the keys `soc`, `criteria` and
        `greg`, where the first one correspond with the SOC value, the second
        is a boolean with True if the electrode material is classified as a
        fast charging one and False if is not and `greg` with the
        `galpynostatic.model.GalvanostaticRegregssor` to allow further
        predictions and plots.

    **kwargs
        Additional keyword arguments that are passed and are documented in
        `galpynostatic.model.GalvanostaticRegressor`.

    Returns
    -------
    soc : float
        BMXFC value.

    res : dict, optional
        A dict present if `full_output=True` and described there.

    References
    ----------
    .. [2] F. Fernandez, E. M. Gavilán-Arriazu, D. E. Barraco, Y. Ein-Eli and
       E. P. M. Leiva. "A metric for benchmarking an extreme fast-charging of
       Li-ion battery electrode materials." `Journal TODO`.
    """
    if isinstance(greg, dict):
        greg_ = GalvanostaticRegressor(d=greg["d"], **kwargs)
        greg_._validate_geometry()
        greg_._map = MapSpline(greg_.dataset)
        greg_.dcoeff_, greg_.k0_ = greg["dcoeff_"], greg["k0_"]
        greg_.dcoeff_err_ = None
    else:
        greg_ = greg

    soc = greg_.predict(np.reshape([60.0 / minutes], (-1, 1)))[0]

    return (
        soc
        if not full_output
        else {"soc": soc, "criteria": soc >= loaded, "greg": greg_}
    )


def fom(d, dcoeff):
    r"""Figure-of-Merit (FOM) for fast charging comparisons.

    This metric was proposed by Xia et al. [3]_ and combines the diffusion
    coefficient, :math:`D`, and the geometric size, :math:`d`, to define the
    characteristic time of diffusion, :math:`\tau`.

    Parameters
    ----------
    d : float
        Geometric size in :math:`cm`.

    dceoff : float
        Diffusion coefficient in :math:`cm^2/s`.

    Returns
    -------
    float
        The FOM characteristic diffusion time (:math:`\tau`) value.

    References
    ----------
    .. [3] H. Xia, W. Zhang, S. Cao and X. Chen. "A figure of merit for
       fast-charging Li-ion battery materials." `ACS Nano, 16` (2022):
       8525-8530.
    """
    return d**2 / dcoeff
